@@ -1,12 +1,16 @@
 /* Keyboard library functions */
 /* H. Hanemaayer (hhanemaa@cs.ruu.nl) */
 
-/* May 28 1998: Changed __keyboard_eventhandler to
- __svgalib_keyboard_eventhandler and made non-static to allow for fake
- keyboard events for the wheelmice (Brion Vibber <brion@pobox.com>) */
-
 /* Apr 03 1998: Added fake mouse event and some cleanup by 101 (Attila Lendvai)
-					e-mail: 101@kempelen.inf.bme.hu */
+        				e-mail: 101@kempelen.inf.bme.hu */
+/*
+ * May 28 1998: Changed __keyboard_eventhandler to
+ * __svgalib_keyboard_eventhandler and made non-static to allow for fake
+ * keyboard events for the wheelmice (Brion Vibber <brion@pobox.com>)
+ *
+ * July 3 1998: Added keyboard remapping support - brion
+*/
+
 
 /*
  * Keyboard I/O based on showkey.c from kbd-0.84 package.
@@ -30,6 +34,10 @@
 /* a few KERNEL ifdefs around kernel data structures though. */
 #include <linux/keyboard.h>
 #include <sys/vt.h>
+/* Needed to check uid of keymap files */
+#include <sys/stat.h>
+#include <unistd.h>
+
 #include <vga.h>
 #include "../libvga.h"
 #include "vgakeyboard.h"
@@ -44,18 +52,47 @@ int __svgalib_kbd_fd = -1; /* nowadays merely used as a flag */
 static int c_state, ctrl_state, alt_state, functionkey_state;
 static int translatemode = 0;
 static unsigned char state[NR_KEYS];	/* NR_KEYS is defined in linux/keyboard.h */
-
+static int keymap[NR_KEYS];
+static int usekeymap = 0;    /* If nonzero, we translate scancodes */
+static int rootkeymaps = 0;  /* If nonzero, only load keymaps owned by root */
+static char keynames[NR_KEYS][MAX_KEYNAME_LEN] = {
+    /* The default US QWERTY layout */
+    "",
+    "Escape", "one", "two", "three", "four", "five", "six", "seven", "eight", "nine", "zero", "minus", "equal", "Delete",
+    "Tab", "q", "w", "e", "r", "t", "y", "u", "i", "o", "p", "bracketleft", "bracketright", "Return",
+    "Control", "a", "s", "d", "f", "g", "h", "j", "k", "l", "semicolon", "apostrophe", "grave",
+    "Shift", "backslash", "z", "x", "c", "v", "b", "n", "m", "comma", "period", "slash", "Shift", "KP_Multiply",
+    "Alt", "space", "Caps_Lock",
+    "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10",
+    "Num_Lock", "Scroll_Lock",
+    "KP_7", "KP_8", "KP_9", "KP_Subtract",
+    "KP_4", "KP_5", "KP_6", "KP_Add",
+    "KP_1", "KP_2", "KP_3",
+    "KP_0", "KP_Period",
+    "Last_Console", "", "less", "F11", "F12", "", "", "", "", "", "", "",
+    "KP_Enter", "Control", "KP_Divide", "Control_backslash", "AltGr", "Break",
+    "Find", "Up", "Prior", "Left", "Right", "Select", "Down", "Next", "Insert",
+    "Remove", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", ""
+};
 static void default_handler(int, int);
+static char *kbd_load_keymap(char *ptr);
+static void kbd_update_keymap(void);
+static int kbd_mapkey(int inscancode);
 
 static struct FakeMouseEvent {
   int	data;
   short type;
   short flags;
+  int   scancode;
+  char  keyname[MAX_KEYNAME_LEN];
 } **fake_mouse_events = 0;
 
 /* Number of events per scancode */
 static short *fme_numberof = 0;
 
+/* Number of scancodes used */
+static int fme_used = 0;
+
 /* Fake event flags */
 #define FMEF_TRIGGERED	1	/* Triggered, */
 #define FMEF_AT_PRESS	2	/* Request at press (or if false at release)*/
@@ -73,10 +110,14 @@ static short *fme_numberof = 0;
 #define FME_TYPE_BUTTON2 9
 #define FME_TYPE_BUTTON3 10
 
-static int expand_events(int scancode) {
- struct FakeMouseEvent *newptr = 0;
+static int expand_events(int n) {
+    struct FakeMouseEvent *newptr = 0;
+//    int n;
 
   if ( ! fake_mouse_events) {
+#ifdef DEBUG_KEYBOARD
+      fprintf(stderr," Allocating space for FMEs...\n");
+#endif
     if ( (fake_mouse_events = malloc(sizeof(void *) * NR_KEYS)) == 0) {
  error:
       fputs("svgalib: keyboard-config: out of memory ?! Fake mouse events might be disabled !", stderr);
@@ -91,9 +132,15 @@ static int expand_events(int scancode) {
       memset(fme_numberof, 0, sizeof(short) * NR_KEYS);
     }
   }
-  if ( (newptr = realloc(fake_mouse_events[scancode],
-	sizeof(struct FakeMouseEvent) * ++fme_numberof[scancode])) != 0 ) {
-    fake_mouse_events[scancode] = newptr;
+#ifdef DEBUG_KEYBOARD
+  fprintf(stderr," Expanding FME #%d... ", n);
+#endif
+  if ( (newptr = realloc(fake_mouse_events[n],
+                         sizeof(struct FakeMouseEvent) * ++fme_numberof[n])) != 0 ) {
+#ifdef DEBUG_KEYBOARD
+  fprintf(stderr,"it's all good.\n");
+#endif
+    fake_mouse_events[n] = newptr;
     return 1;
   } else {
     goto error;
@@ -101,7 +148,7 @@ static int expand_events(int scancode) {
 }
 
 static char *kbd_config_options[] = {
-  "kbd_fake_mouse_event", NULL
+  "kbd_fake_mouse_event", "kbd_keymap", "kbd_only_root_keymaps", NULL
 };
 
 static char *kbd_process_option(int option, int mode) {
@@ -112,21 +159,37 @@ static char *kbd_process_option(int option, int mode) {
   int data     = 0;
   int scancode = 0;
   int event_ok = 0;
+  int n = 0;
   struct FakeMouseEvent *event;
+  char keyname[MAX_KEYNAME_LEN];
   
   switch (option) {
-    case 0:
+    case 0: /* kbd_fake_mouse_event */
 	if ( (ptr = strtok(NULL, " ")) == 0) {
   param_needed:
 	  fprintf(stderr, "svgalib: kbd-config: too few parameters for \'%s\'\n",
 	  		kbd_config_options[option]);
 	  return ptr;
-	} else if ( (scancode = atoi(ptr)) == 0 || scancode & 0x80
-			|| abs(scancode) > NR_KEYS) {
+        } else if ( (scancode = __svgalib_mapkeyname(ptr)) == -1) {
 	  fprintf(stderr, "svgalib: kbd-config: \'%s\' is not a valid scancode\n", ptr);
 	  return ptr;
-	}
+        }
 
+        strncpy(keyname, ptr, MAX_KEYNAME_LEN);
+
+        /* Find the event for this key if there is one*/
+#ifdef DEBUG_KEYBOARD
+        fprintf(stderr," Looking for fme for key \'%s\'... ", keyname);
+#endif
+        for(n = 0; n < fme_used; n++)
+            if(strcmp(fake_mouse_events[n]->keyname, keyname) == 0)
+                break;
+        if(n == fme_used)
+            fme_used++;
+#ifdef DEBUG_KEYBOARD
+        fprintf(stderr," found at #%d.\n", n);
+#endif
+        
   read_event:
 	if ( (ptr = strtok(NULL, " ")) == 0) {
 	  if ( event_ok )
@@ -166,12 +229,14 @@ static char *kbd_process_option(int option, int mode) {
 #ifdef DEBUG_KEYBOARD
   printf(" Fake Mouse Event: scancode: %ld;  type: %ld;  data: %ld;  flags: %ld\n", (long)scancode, (long)type, (long)data, (long)flags);
 #endif
-	   if ( ! expand_events(scancode))
+	   if ( ! expand_events(n))
 	     return ptr;
-	   event = &fake_mouse_events[scancode][fme_numberof[scancode] - 1];
+           event = &fake_mouse_events[n][fme_numberof[n] - 1];
 	   event -> type  = type;
 	   event -> flags = flags;
 	   event -> data  = data;
+           event -> scancode = scancode;
+           strcpy(event -> keyname, keyname);
 	   event_ok = 1;
 	   goto read_event;
 	} else if ( ! strcasecmp(ptr, "deltax")) {
@@ -229,12 +294,29 @@ static char *kbd_process_option(int option, int mode) {
 	} else
 	  return ptr;
     	break;
-  
+
+    case 1: /* kbd_keymap */
+         if ( (ptr = strtok(NULL, " ")) == 0 ) {
+             goto param_needed;
+         } else
+             if(kbd_load_keymap(ptr))
+                 return ptr;
+
+         break;
+
+  case 2: /* kbd_only_root_keymaps */
+#ifdef DEBUG_KEYBOARD
+      fprintf(stderr, " Setting rootkeymaps to 1.\n");
+#endif
+      rootkeymaps = 1;
+      break;
+
   }
   return strtok(NULL, " ");
 };
 
 int keyboard_init_return_fd(void) {
+    char *ptr;
 
     keyboard_translatekeys(translatemode); /* Honour 'nosigint' setting */
 
@@ -264,6 +346,11 @@ int keyboard_init_return_fd(void) {
 
     __svgalib_read_options(kbd_config_options, kbd_process_option);
 
+    /* Check SVGALIB_KEYMAP env var */
+    if ( (ptr = getenv("SVGALIB_KEYMAP")) != 0) {
+        kbd_load_keymap(ptr);
+    }
+
     return __svgalib_kbd_fd;		/* OK, return fd. */
 }
 
@@ -371,7 +458,7 @@ static int keyboard_getevents(int wait)
 	eventhandled = 1;
 
     for (i = 0; i < bytesread; i++) {
-      unsigned char scancode = buf[i] & 0x7f;
+      unsigned char scancode = kbd_mapkey(buf[i] & 0x7f);
       unsigned char is_pressed = (buf[i] & 0x80) ? 0 : 1;
 
 	/* Check for ctrl-c. */
@@ -397,79 +484,87 @@ static int keyboard_getevents(int wait)
 		break;
 	}
 
-	if (fake_mouse_events  &&  fme_numberof[scancode]
-						&& __svgalib_mouse_fd > -1) {
-	 int i;
-	 int dx=0, dy=0, dz=0, but=0; /* To gather mutiple events into one */
-	 int but_changed = 0;
-	 struct FakeMouseEvent *event;
+	if (fake_mouse_events  && __svgalib_mouse_fd > -1) {
+            int n;
 
-	  for (i = 0; i < fme_numberof[scancode]; i++) {
-	   int request_at_down;
-	   short flags;
-	    event = &fake_mouse_events[scancode][i];
-	    if (!is_pressed)
-	      event -> flags &= ~FMEF_TRIGGERED;
-	    flags = event -> flags;
-	    request_at_down = (event -> flags & FMEF_AT_PRESS)  ?  1 : 0;
-							#ifdef DEBUG_KEYBOARD
-							  printf(" event type: %ld;  flags: %ld;  data: %ld; is_pressed: %ld\n", (long)event->type, (long)flags, (long)event->data, (long)is_pressed);
-							#endif
-	    if (flags & FMEF_AT_BOTH  ||  request_at_down == is_pressed) {
-							#ifdef DEBUG_KEYBOARD
-							  printf("   flags: %ld\n", (long)flags);
-							#endif 
-	      if (  (! (flags & FMEF_TRIGGERED))  ||  flags & FMEF_REPEAT) {
-							#ifdef DEBUG_KEYBOARD
-							  printf("      triggering\n");
-							#endif
-		switch (event -> type) {
-		  case FME_TYPE_BUTTON1:
-		    but = (event -> data) ?  but | 4  :  but & ~4;
-		    but_changed = 1;
-		    break;
-		  case FME_TYPE_BUTTON2:
-		    but = (event -> data) ?  but | 1  :  but & ~1;
-		    but_changed = 1;
-		    break;
-		  case FME_TYPE_BUTTON3:
-		    but = (event -> data) ?  but | 2  :  but & ~2;
-		    but_changed = 1;
-		    break;
+            for(n = 0; n < fme_used; n++)
+                if(fake_mouse_events[n][0].scancode == scancode)
+                    break;
 
-		  case FME_TYPE_IGNOREX:
-		    __svgalib_m_ignore_dx = event -> data;
-		    break;
-		  case FME_TYPE_IGNOREY:
-		    __svgalib_m_ignore_dy = event -> data;
-		    break;
-		  case FME_TYPE_IGNOREZ:
-		    __svgalib_m_ignore_dz = event -> data;
-		    break;
+            if(fme_numberof[n])
+            {
+                int i;
+                int dx=0, dy=0, dz=0, but=0; /* To gather mutiple events into one */
+                int but_changed = 0;
+                struct FakeMouseEvent *event;
 
-		  case FME_TYPE_DELTAX:
-		    dx += event -> data;
-		    break;
-		  case FME_TYPE_DELTAY:
-		    dy += event -> data;
-		    break;
-		  case FME_TYPE_DELTAZ:
-		    dz += event -> data;
-		    break;
-		}
-		if (is_pressed)
-		  event -> flags |= FMEF_TRIGGERED;
-	      }
-	    }
-	  }
-	  if ((dx || dy || dz || but_changed) && __svgalib_mouse_eventhandler) {
-	    __svgalib_mouse_eventhandler(but, dx, dy, dz, 0, 0, 0);
+                for (i = 0; i < fme_numberof[n]; i++) {
+                    int request_at_down;
+                    short flags;
+                    event = &fake_mouse_events[n][i];
+                    if (!is_pressed)
+                        event -> flags &= ~FMEF_TRIGGERED;
+                    flags = event -> flags;
+                    request_at_down = (event -> flags & FMEF_AT_PRESS)  ?  1 : 0;
+#ifdef DEBUG_KEYBOARD
+                    printf(" event type: %ld;  flags: %ld;  data: %ld; is_pressed: %ld\n", (long)event->type, (long)flags, (long)event->data, (long)is_pressed);
+#endif
+                    if (flags & FMEF_AT_BOTH  ||  request_at_down == is_pressed) {
+#ifdef DEBUG_KEYBOARD
+                        printf("   flags: %ld\n", (long)flags);
+#endif
+                        if (  (! (flags & FMEF_TRIGGERED))  ||  flags & FMEF_REPEAT) {
+#ifdef DEBUG_KEYBOARD
+                            printf("      triggering\n");
+#endif
+                            switch (event -> type) {
+                            case FME_TYPE_BUTTON1:
+                                but = (event -> data) ?  but | 4  :  but & ~4;
+                                but_changed = 1;
+                                break;
+                            case FME_TYPE_BUTTON2:
+                                but = (event -> data) ?  but | 1  :  but & ~1;
+                                but_changed = 1;
+                                break;
+                            case FME_TYPE_BUTTON3:
+                                but = (event -> data) ?  but | 2  :  but & ~2;
+                                but_changed = 1;
+                                break;
+
+                            case FME_TYPE_IGNOREX:
+                                __svgalib_m_ignore_dx = event -> data;
+                                break;
+                            case FME_TYPE_IGNOREY:
+                                __svgalib_m_ignore_dy = event -> data;
+                                break;
+                            case FME_TYPE_IGNOREZ:
+                                __svgalib_m_ignore_dz = event -> data;
+                                break;
+
+                            case FME_TYPE_DELTAX:
+                                dx += event -> data;
+                                break;
+                            case FME_TYPE_DELTAY:
+                                dy += event -> data;
+                                break;
+                            case FME_TYPE_DELTAZ:
+                                dz += event -> data;
+                                break;
+                            }
+                            if (is_pressed)
+                                event -> flags |= FMEF_TRIGGERED;
+                        }
+                    }
+                }
+                if ((dx || dy || dz || but_changed) && __svgalib_mouse_eventhandler) {
+              __svgalib_mouse_eventhandler(but, dx, dy, dz, 0, 0, 0);
 
 #ifdef DEBUG_KEYBOARD
-	    printf("\tfake_mouse_event triggered; but: %ld; dx: %ld; dy: %ld; dz: %ld\n",
-	    		(long)but, (long)dx, (long)dy, (long)dz);
+            printf("\tfake_mouse_event triggered; but: %ld; dx: %ld; dy: %ld; dz: %ld\n",
+                   (long)but, (long)dx, (long)dy, (long)dz);
 #endif
-	  }
+          }
+            }
 	}
 
 	if (ctrl_state && c_state && !(translatemode & DONT_CATCH_CTRLC))
@@ -499,7 +594,7 @@ static int keyboard_getevents(int wait)
 	        return 1;
 	    }
 	}
-	__svgalib_keyboard_eventhandler(buf[i] & 0x7f,
+	__svgalib_keyboard_eventhandler(scancode,
 		    (buf[i] & 0x80) ? KEY_EVENTRELEASE : KEY_EVENTPRESS);
     }
 
@@ -623,3 +718,162 @@ void keyboard_translatekeys(int mode)
     if (__svgalib_nosigint)
 	translatemode |= DONT_CATCH_CTRLC;
 }
+
+static int kbd_mapkey(int inscancode)
+{
+    if (usekeymap)
+        return keymap[inscancode];
+    else
+        return inscancode;
+}
+
+static char *kbd_load_keymap(char *ptr)
+{
+    /* Find the specified keymap file */
+    /* For now, use a full pathname */
+    FILE *keymapfile;
+
+#ifdef DEBUG_KEYBOARD
+    fprintf(stderr, " Trying to load keymap \'%s\'...\n",ptr);
+#endif
+
+    /* If so configured check if the file is owned by root */
+    if(rootkeymaps) {
+        struct stat fs;
+        if(stat(ptr, &fs)) {
+            fprintf(stderr, "svgalib: kbd-config: cannot stat keymap file \'%s\'\n",
+                    ptr);
+            return ptr;
+        } else if(fs.st_uid != 0) {
+            fprintf(stderr, "svgalib: kbd-config: keymap file \'%s\' not owned by root\n",
+                    ptr);
+            return ptr;
+        }
+    }
+
+    if((keymapfile = fopen(ptr, "rt"))) {
+        /* Load it in! */
+        char buf[81], nbuf[81];
+        int i, l = 0, insc, outsc;
+
+        buf[80] = nbuf[80] = 0; /* Protect somewhat against overruns */
+
+        /* Disable the keymap until we're done in case something goes wrong... */
+        usekeymap = 0;
+
+        /* Initialize keymap */
+        for (i = 0; i < NR_KEYS; keymap[i++] = i);
+
+        while(!feof(keymapfile)) {
+            /* Read */
+            fgets(buf, 80, keymapfile);
+            l++;
+
+            /* Ignore comments & blank lines */
+            if ((buf[0] != '#') && (buf[0] != '\n')) {
+                /* Interpret the numbers */
+                if((i = sscanf(buf, "%d %d %s", &insc, &outsc, nbuf) == 3)) {
+                    if(checkscancode(insc)) continue;
+                    if(checkscancode(outsc)) continue;
+                    keymap[insc] = outsc;
+                    strncpy(keynames[insc], nbuf, MAX_KEYNAME_LEN);
+                } else {
+                    fprintf(stderr, "svgalib: kbd-config: skipping line %d of keymap - bad %sput scancode\n",
+                            l, ((i == 1)?("out"):("in")));
+                }
+            }
+        }
+
+        usekeymap = 1;
+
+        /* Update the mouse driver's fake keyboard events */
+#ifdef DEBUG_KEYBOARD
+        fprintf(stderr, " Keymap loaded. Updating fake keyboard events...\n");
+#endif
+        __svgalib_mouse_update_keymap();
+
+        /* And the fake mouse events */
+#ifdef DEBUG_KEYBOARD
+        fprintf(stderr, " Updating fake mouse events...\n");
+#endif
+        kbd_update_keymap();
+
+        fclose(keymapfile);
+    } else {
+        fprintf(stderr, "svgalib: kbd-config: keymap file \'%s\' cannot be opened\n", ptr);
+        return ptr;
+    }
+
+    return NULL;
+}
+
+int __svgalib_mapkeyname(const char *keyname)
+{
+    int scancode;
+    char *ptr;
+
+#ifdef DEBUG_KEYBOARD
+    fprintf(stderr, " Attempting to map \'%s\' to a scancode...", keyname);
+#endif
+    
+    if(!keyname) {
+        fprintf(stderr, "svgalib: kbd-config: can't use NULL keyname!\n");
+        return -1;
+    }
+
+    if(!*keyname) {
+        fprintf(stderr, "svgalib: kbd-config: can't use empty keyname!\n");
+        return -1;
+    }
+
+    /* Is it a number? */
+    scancode = strtol(keyname, &ptr, 0);
+    if(ptr != keyname) {
+        /* Is it in range? */
+        if((scancode < 0) || (scancode >= NR_KEYS)) {
+            fprintf(stderr, "svgalib: kbd-config: scancode %s out of range!\n",keyname);
+            return -1;
+        } else {
+#ifdef DEBUG_KEYBOARD
+            fprintf(stderr, " mapped numerically to %d.\n", scancode);
+#endif
+            return scancode;
+        }
+    }
+
+    /* Look up a symbolic name */
+    for(scancode = 0; scancode < NR_KEYS; scancode++) {
+#ifdef DEBUG_KEYBOARD
+        fprintf(stderr, " (%s)", keynames[scancode]);
+#endif
+        if(strncasecmp(keyname, keynames[scancode], MAX_KEYNAME_LEN) == 0) {
+#ifdef DEBUG_KEYBOARD
+            fprintf(stderr, " mapped symbolically to %d", scancode);
+            if(usekeymap)
+                fprintf(stderr, ", remapped to %d", kbd_mapkey(scancode));
+            fprintf(stderr, ".\n");
+#endif
+            return kbd_mapkey(scancode);
+        }
+    }
+
+#ifdef DEBUG_KEYBOARD
+    fprintf(stderr, " no match!\n");
+#endif
+    
+    /* No match */
+    return -1;
+}
+
+static void kbd_update_keymap(void)
+{
+    /* Remap the scancodes of the fake mouse events */
+    int n, i=0;
+    
+    for(n = 0; n < fme_used; n++)
+        if(fake_mouse_events[n]) {
+            /*for(i = 0; i < fme_numberof[n]; i++)*/
+                fake_mouse_events[n][i].scancode =
+                    __svgalib_mapkeyname(fake_mouse_events[n][i].keyname);
+        }
+}
