@@ -672,7 +672,8 @@ DriverSpecs __svgalib_tvga8900_driverspecs =
     0,				/* extset */
     0,
     0,				/* linear */
-    NULL			/* accelspecs */
+    NULL,                       /* Accelspecs */
+    NULL,                       /* Emulation */
 };
 
 
